==> masamune_firebase-0.1.3.tar.gz: lib/firestore/firestorecollection.dart <==
part of masamune.firebase;

/// Get the Firestore collection.
/// Normally retrieve 200 documents from the path, but you can change the search criteria by specifying [query]
/// ```
/// FirestoreCollection col = await FirestoreCollection.listen( "user/user" );
/// for( FirestoreDocument doc in col ) { ... }
/// String name = doc.getString( "name" );
/// newDocument["age"] = 18;
/// newDocument.save();
/// ```
class FirestoreCollection extends TaskCollection<FirestoreDocument> implements ITask, IDataCollection<FirestoreDocument>, IFirestoreChangeListener, IFirestoreCollection {
  /// Create a Completer that matches the class.
  /// Do not use from external class
  @override
  @protected
  Completer createCompleter() => Completer<FirestoreCollection>();
  /// Process to create a new instance
  /// Do not use from outside the class
  @override
  @protected
  T createInstance<T extends IClonable>() => FirestoreCollection._( ) as T;
  Firebase get _app {
    if( this.__app == null ) this.__app = Firebase._find(this.protocol);
    return this.__app;
  }
  Firebase __app;
  FirestoreAuth get _auth {
    if( this.__auth == null ) this.__auth = FirestoreAuth._find(this.protocol);
    return this.__auth;
  }
  FirestoreAuth __auth;
  Query get _reference {
    if( this.__reference == null && this.rawPath != null && isNotEmpty( this.rawPath.path ) ) this.__reference = this._app._db.collection( this.rawPath.path);
    return this.__reference;
  }
  Query __reference;
  StreamSubscription<QuerySnapshot> _listener;

  /// Get the Firestore collection.
  /// Normally retrieve 200 documents from the path, but you can change the search criteria by specifying [query]
  /// ```
  /// FirestoreCollection col = await FirestoreCollection.listen( "user/user" );
  /// for( FirestoreDocument doc in col ) { ... }
  /// String name = doc.getString( "name" );
  /// newDocument["age"] = 18;
  /// newDocument.save();
  /// ```
  /// [path]: Collection path
  /// [query]: Querying collections
  static Future<FirestoreCollection> listen( String path, {FirestoreQuery query, OrderBy orderBy = OrderBy.None, OrderBy thenBy = OrderBy.None, String orderByKey, String thenByKey} ){  
    path = path?.trimString( Const.slash )?.replaceAll( "//", "/" )?.applyTags();
    if( isEmpty(path) ) {
      Log.error( "Path is invalid." );
      return null;
    }
    int length = Paths.length( path );
    if( length <= 0 || length % 2 != 1 ) {
      Log.error( "Path is not document path." );
      return null;
    }
    FirestoreCollection collection = PathMap.get<FirestoreCollection>( path );
    if( collection != null ) {
      if( query != null ) collection.query = query;
      if( orderBy != OrderBy.None ) collection.orderBy = orderBy;
      if( thenBy != OrderBy.None ) collection.thenBy = thenBy;
      if( isNotEmpty(orderByKey) ) collection.orderByKey = orderByKey;
      if( isNotEmpty(thenByKey) ) collection.thenByKey = thenByKey;
      return collection.reload( );
    }
    collection = FirestoreCollection._( path:path );
    if( query != null ) collection.query = query;
    if( orderBy != OrderBy.None ) collection.orderBy = orderBy;
    if( thenBy != OrderBy.None ) collection.thenBy = thenBy;
    if( isNotEmpty(orderByKey) ) collection.orderByKey = orderByKey;
    if( isNotEmpty(thenByKey) ) collection.thenByKey = thenByKey;
    collection._constructListener( );
    return collection.future;
  }
  FirestoreCollection._({String path, Iterable<FirestoreDocument> children, int group = 0, int order = 10}) : super( path:path, children:children, group:group, order:order );


  /// Update document data
  Future<T> reload<T extends IFirestoreChangeListener>( ){
    this.init();
    this._constructListener();
    return this.future;
  }  
  void _constructListener( ) async {
    try {
      if( this._listener != null ) {
          await this._listener.cancel();
          this._listener = null;
      }
      if( this._app == null ) this.__app = await Firebase.initialize();
      if( this._auth == null ) this.__auth = await FirestoreAuth.signIn(protocol:this.protocol);
      this.__reference = this._buildQueryInternal( this._app._db.collection( this.rawPath.path ) );
      this._listener = this._reference.snapshots().listen(
        ( snapshot ) {
          Map<String, Map<String, dynamic>> data = MapPool.get();
          for( DocumentSnapshot doc in snapshot.documents ){
            if( doc == null || !doc.exists ) continue;
            data[doc.documentID] = doc.data;
          }
          this._done(data, DateTime.now().frameMillisecondsSinceEpoch);
        }
      );
    } catch( e ) {
      this.error( e.toString() );
      this.dispose( );
    }
  }
  void _done( Map<String, Map<String, dynamic>> data, int updatedTime ) {
    if( this.isDone && this.length== data.length && updatedTime <= this.updatedTime ) return;
    if( this.isUpdating ) return;
    if( data != null ) {
      List<FirestoreDocument> addData = ListPool.get();
      data.forEach(
        (key, value) {
          if( isEmpty( key ) || value == null ) return;
          FirestoreDocument doc = this.data.firstWhere( ( k, v ) => v?.id == key, orElse: () => null );
          if( doc != null ) {
            doc._setInternal( value );
          } else {
            addData.add( FirestoreDocument._create( Paths.child( this.path, key ), value ) );
          }
          value.release();
        }
      );
      this._setInternal( addData );
      for( int i = this.data.length - 1; i >= 0; i-- ){
        FirestoreDocument doc = this.data[i];
        if( doc == null ) continue;
        if( !data.containsKey( doc.id ) ) this.remove( doc );
      }
      Log.msg( "Updated data: %s (%s)", [this.path, this.runtimeType] );
    }
    this.data.sort(this._sortAction);
    this.done( );
    data.release( );
  }
  void _setInternal( Iterable<FirestoreDocument> children ) {
    if( children != null ) {
      for( FirestoreDocument doc in children ) {
        if( doc == null ) continue;
        this.setInternal(doc);
      }
    }
  }
  
  
  /// Apply the changed collection data to the remote Firestore
  /// It may take up to 1 second or more
  /// Use the [isUpdating] flag if you want to check if it is updating
  void save( ) {
    for( FirestoreDocument tmp in this.data.values ) {
      if( tmp == null ) continue;
      tmp.save( );
    }
  }
  /// Delete this FirestoreCollection path from the remote Firestore.
  /// This object itself is also [dispose()]
  /// It may take up to 1 second or more
  /// Use the [isUpdating] flag if you want to check if it is updating
  void delete( ) {
    for( int i = this.data.length - 1; i >= 0; i-- ){
      FirestoreDocument doc = this.data[i];
      if( doc == null ) continue;
      doc.delete( );
      this.data.removeAt( i );
    }
  }
  Query _buildQueryInternal( CollectionReference reference ) {
    if( this.query == null || reference == null || isEmpty( this.query.key ) ) return reference;
    Query fquery = reference;
    switch( this.query.type ) {
      case FirestoreQueryType.EqualTo:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, isEqualTo:this.query.value );
      break;
      case FirestoreQueryType.HigherThan:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, isGreaterThanOrEqualTo:this.query.value );
      break;
      case FirestoreQueryType.LowerThan:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, isLessThanOrEqualTo:this.query.value );
      break;
      case FirestoreQueryType.Range:
      if( this.query.value is Range ) fquery = fquery.where( this.query.key, isGreaterThanOrEqualTo:this.query.value.min ).where( this.query.key, isLessThanOrEqualTo:this.query.value.max );
      break;
      case FirestoreQueryType.ArrayContains:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, arrayContains: this.query.value );
      break;
      case FirestoreQueryType.ArrayContainsAny:
      if( this.query.contains != null && this.query.contains.length > 0 ) fquery = fquery.where( this.query.key, arrayContainsAny:this.query.contains );
      break;
      case FirestoreQueryType.In:
      if( this.query.contains != null && this.query.contains.length> 0 ) fquery = fquery.where( this.query.key, whereIn:this.query.contains );
      break;
      default:
      break;
    }
    switch( this.orderBy ) {
      case OrderBy.Asc:
      fquery = fquery.orderBy( this.query.key ).startAt( [this.query.index * this.query.limit] ).limit( this.query.limit );
      break;
      case OrderBy.Desc:
      fquery = fquery.orderBy( this.query.key, descending:true ).startAt( [this.query.index * this.query.limit] ).limit( this.query.limit );
      break;
      default:
      break;
    }
    return fquery;
  }
  /// Set the query
  /// After setting the query, execute [reload()] to reflect the setting.
  FirestoreQuery query = FirestoreQuery( );

  /// Order for sorting
  OrderBy orderBy;
  /// Compare key for sorting
  String orderByKey;
  /// Order for sorting
  /// Apply when the first sort is equivalent
  OrderBy thenBy;
  /// Compare key for sorting
  /// Apply when the first sort is equivalent
  String thenByKey;

  /// Sort according to the sort setting
  /// [orderBy] and [orderByKey], [thenBy] and [thenByKey] as the sort contents.
  void sort( ) => this.data.sort( this._sortAction );
  int _sortAction( MapEntry<String, FirestoreDocument> a, MapEntry<String, FirestoreDocument> b ) {
    int r = 0;
    if( a.value != null && b.value != null && a.value.containsKey( this.orderByKey ) && b.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        if( a.value[this.orderByKey] is double ) {
          r = a.value.getDouble( this.orderByKey ).compareTo( b.value.getDouble( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is int ) {
          r = a.value.getInt( this.orderByKey ).compareTo( b.value.getInt( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is bool ) {
          r = a.value.getBool( this.orderByKey ).compareTo( b.value.getBool( this.orderByKey ) );
        } else {
          r = a.value.getString( this.orderByKey ).compareTo( b.value.getString( this.orderByKey ) );
        }
      } else {
        if( a.value[this.orderByKey] is double ) {
          r = b.value.getDouble( this.orderByKey ).compareTo( a.value.getDouble( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is int ) {
          r = b.value.getInt( this.orderByKey ).compareTo( a.value.getInt( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is bool ) {
          r = b.value.getBool( this.orderByKey ).compareTo( a.value.getBool( this.orderByKey ) );
        } else {
          r = b.value.getString( this.orderByKey ).compareTo( a.value.getString( this.orderByKey ) );
        }
      }
    } else if( a.value == null || !a.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        r = -1;
      } else {
        r = 1;
      }
    } else if( b.value == null || !b.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        r = 1;
      } else {
        r = -1;
      }
    }
    if( r == 0 && this.thenBy != OrderBy.None && !isEmpty( this.thenByKey ) ) {
      if( a.value != null && b.value != null && a.value.containsKey( this.thenByKey ) && b.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          if( a.value[this.thenByKey] is double ) {
            r = a.value.getDouble( this.thenByKey ).compareTo( b.value.getDouble( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is int ) {
            r = a.value.getInt( this.thenByKey ).compareTo( b.value.getInt( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is bool ) {
            r = a.value.getBool( this.thenByKey ).compareTo( b.value.getBool( this.thenByKey ) );
          } else {
            r = a.value.getString( this.thenByKey ).compareTo( b.value.getString( this.thenByKey ) );
          }
        } else {
          if( a.value[this.thenByKey] is double ) {
            r = b.value.getDouble( this.thenByKey ).compareTo( a.value.getDouble( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is int ) {
            r = b.value.getInt( this.thenByKey ).compareTo( a.value.getInt( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is bool ) {
            r = b.value.getBool( this.thenByKey ).compareTo( a.value.getBool( this.thenByKey ) );
          } else {
            r = b.value.getString( this.thenByKey ).compareTo( a.value.getString( this.thenByKey ) );
          }
        }
      } else if( a.value == null || !a.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          r = -1;
        } else {
          r = 1;
        }
      } else if( b.value == null || !b.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          r = 1;
        } else {
          r = -1;
        }
      }
    }
    return r;
  }

  /// True if communicating with the server for saving or deleting
  bool get isUpdating {
    for( FirestoreDocument doc in this.data.values ){
      if( doc == null ) continue;
      if( doc.isUpdating ) return true;
    }
    return false;
  }

  
  /// Get the protocol of the path
  @override
  String get protocol {
    if( isEmpty( this.rawPath.scheme ) ) return "firestore";
    return this.rawPath.scheme;
  }
  /// Destroys the object
  /// 
  /// Destroyed objects are not allowed
  @override  
  void dispose( ) {
    super.dispose( );
    this._disposeInternal( );
  }
  void _disposeInternal( ) {
    this.__reference = null;
    if( this._listener == null ) return;
    this._listener.cancel( );
    this._listener = null;
  }
  /// Callback event when application quit
  void onApplicationQuit( ) {
    super.onApplicationQuit( );
    this._disposeInternal( );
  }

}

==> masamune_firebase-0.1.4.tar.gz: lib/firestore/firestorecollection.dart <==
part of masamune.firebase;

/// Get the Firestore collection.
/// Normally retrieve 200 documents from the path, but you can change the search criteria by specifying [query]
/// ```
/// FirestoreCollection col = await FirestoreCollection.listen( "user/user" );
/// for( FirestoreDocument doc in col ) { ... }
/// String name = doc.getString( "name" );
/// newDocument["age"] = 18;
/// newDocument.save();
/// ```
class FirestoreCollection extends TaskCollection<FirestoreDocument> implements ITask, IDataCollection<FirestoreDocument>, IFirestoreChangeListener, IFirestoreCollection {
  /// Create a Completer that matches the class.
  /// Do not use from external class
  @override
  @protected
  Completer createCompleter() => Completer<FirestoreCollection>();
  /// Process to create a new instance
  /// Do not use from outside the class
  @override
  @protected
  T createInstance<T extends IClonable>() => FirestoreCollection._( ) as T;
  Firebase get _app {
    if( this.__app == null ) this.__app = Firebase._find(this.protocol);
    return this.__app;
  }
  Firebase __app;
  FirestoreAuth get _auth {
    if( this.__auth == null ) this.__auth = FirestoreAuth._find(this.protocol);
    return this.__auth;
  }
  FirestoreAuth __auth;
  Query get _reference {
    if( this.__reference == null && this.rawPath != null && isNotEmpty( this.rawPath.path ) ) this.__reference = this._app._db.collection( this.rawPath.path);
    return this.__reference;
  }
  Query __reference;
  StreamSubscription<QuerySnapshot> _listener;

  /// Get the Firestore collection.
  /// Normally retrieve 200 documents from the path, but you can change the search criteria by specifying [query]
  /// ```
  /// FirestoreCollection col = await FirestoreCollection.listen( "user/user" );
  /// for( FirestoreDocument doc in col ) { ... }
  /// String name = doc.getString( "name" );
  /// newDocument["age"] = 18;
  /// newDocument.save();
  /// ```
  /// [path]: Collection path
  /// [query]: Querying collections
  static Future<FirestoreCollection> listen( String path, {FirestoreQuery query, OrderBy orderBy = OrderBy.None, OrderBy thenBy = OrderBy.None, String orderByKey, String thenByKey} ){  
    path = path?.trimString( Const.slash )?.replaceAll( "//", "/" )?.applyTags();
    if( isEmpty(path) ) {
      Log.error( "Path is invalid." );
      return null;
    }
    int length = Paths.length( path );
    if( length <= 0 || length % 2 != 1 ) {
      Log.error( "Path is not document path." );
      return null;
    }
    FirestoreCollection collection = PathMap.get<FirestoreCollection>( path );
    if( collection != null ) {
      if( query != null ) collection.query = query;
      if( orderBy != OrderBy.None ) collection.orderBy = orderBy;
      if( thenBy != OrderBy.None ) collection.thenBy = thenBy;
      if( isNotEmpty(orderByKey) ) collection.orderByKey = orderByKey;
      if( isNotEmpty(thenByKey) ) collection.thenByKey = thenByKey;
      return collection.reload( );
    }
    collection = FirestoreCollection._( path:path );
    if( query != null ) collection.query = query;
    if( orderBy != OrderBy.None ) collection.orderBy = orderBy;
    if( thenBy != OrderBy.None ) collection.thenBy = thenBy;
    if( isNotEmpty(orderByKey) ) collection.orderByKey = orderByKey;
    if( isNotEmpty(thenByKey) ) collection.thenByKey = thenByKey;
    collection._constructListener( );
    return collection.future;
  }
  FirestoreCollection._({String path, Iterable<FirestoreDocument> children, int group = 0, int order = 10}) : super( path:path, children:children, group:group, order:order );


  /// Update document data
  Future<T> reload<T extends IFirestoreChangeListener>( ){
    this.init();
    this._constructListener();
    return this.future;
  }  
  void _constructListener( ) async {
    try {
      if( this._listener != null ) {
          await this._listener.cancel();
          this._listener = null;
      }
      if( this._app == null ) this.__app = await Firebase.initialize();
      if( this._auth == null ) this.__auth = await FirestoreAuth.signIn(protocol:this.protocol);
      this.__reference = this._buildQueryInternal( this._app._db.collection( this.rawPath.path ) );
      this._listener = this._reference.snapshots().listen(
        ( snapshot ) {
          Map<String, Map<String, dynamic>> data = MapPool.get();
          for( DocumentSnapshot doc in snapshot.documents ){
            if( doc == null || !doc.exists ) continue;
            data[doc.documentID] = doc.data;
          }
          this._done(data, DateTime.now().frameMillisecondsSinceEpoch);
        }
      );
    } catch( e ) {
      this.error( e.toString() );
      this.dispose( );
    }
  }
  void _done( Map<String, Map<String, dynamic>> data, int updatedTime ) {
    if( this.isDone && this.length== data.length && updatedTime <= this.updatedTime ) return;
    if( this.isUpdating ) return;
    if( data != null ) {
      List<FirestoreDocument> addData = ListPool.get();
      data.forEach(
        (key, value) {
          if( isEmpty( key ) || value == null ) return;
          FirestoreDocument doc = this.data.firstWhere( ( k, v ) => v?.id == key, orElse: () => null );
          if( doc != null ) {
            doc._setInternal( value );
          } else {
            addData.add( FirestoreDocument._create( Paths.child( this.path, key ), value ) );
          }
          value.release();
        }
      );
      this._setInternal( addData );
      for( int i = this.data.length - 1; i >= 0; i-- ){
        FirestoreDocument doc = this.data[i];
        if( doc == null ) continue;
        if( !data.containsKey( doc.id ) ) this.remove( doc );
      }
      Log.ast( "Updated data: %s (%s)", [this.path, this.runtimeType] );
    }
    this.data.sort(this._sortAction);
    this.done( );
    data.release( );
  }
  void _setInternal( Iterable<FirestoreDocument> children ) {
    if( children != null ) {
      for( FirestoreDocument doc in children ) {
        if( doc == null ) continue;
        this.setInternal(doc);
      }
    }
  }
  
  
  /// Apply the changed collection data to the remote Firestore
  /// It may take up to 1 second or more
  /// Use the [isUpdating] flag if you want to check if it is updating
  void save( ) {
    for( FirestoreDocument tmp in this.data.values ) {
      if( tmp == null ) continue;
      tmp.save( );
    }
  }
  /// Delete this FirestoreCollection path from the remote Firestore.
  /// This object itself is also [dispose()]
  /// It may take up to 1 second or more
  /// Use the [isUpdating] flag if you want to check if it is updating
  void delete( ) {
    for( int i = this.data.length - 1; i >= 0; i-- ){
      FirestoreDocument doc = this.data[i];
      if( doc == null ) continue;
      doc.delete( );
      this.data.removeAt( i );
    }
  }
  Query _buildQueryInternal( CollectionReference reference ) {
    if( this.query == null || reference == null || isEmpty( this.query.key ) ) return reference;
    Query fquery = reference;
    switch( this.query.type ) {
      case FirestoreQueryType.EqualTo:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, isEqualTo:this.query.value );
      break;
      case FirestoreQueryType.HigherThan:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, isGreaterThanOrEqualTo:this.query.value );
      break;
      case FirestoreQueryType.LowerThan:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, isLessThanOrEqualTo:this.query.value );
      break;
      case FirestoreQueryType.Range:
      if( this.query.value is Range ) fquery = fquery.where( this.query.key, isGreaterThanOrEqualTo:this.query.value.min ).where( this.query.key, isLessThanOrEqualTo:this.query.value.max );
      break;
      case FirestoreQueryType.ArrayContains:
      if( this.query.value != null ) fquery = fquery.where( this.query.key, arrayContains: this.query.value );
      break;
      case FirestoreQueryType.ArrayContainsAny:
      if( this.query.contains != null && this.query.contains.length > 0 ) fquery = fquery.where( this.query.key, arrayContainsAny:this.query.contains );
      break;
      case FirestoreQueryType.In:
      if( this.query.contains != null && this.query.contains.length> 0 ) fquery = fquery.where( this.query.key, whereIn:this.query.contains );
      break;
      default:
      break;
    }
    switch( this.orderBy ) {
      case OrderBy.Asc:
      fquery = fquery.orderBy( this.query.key ).startAt( [this.query.index * this.query.limit] ).limit( this.query.limit );
      break;
      case OrderBy.Desc:
      fquery = fquery.orderBy( this.query.key, descending:true ).startAt( [this.query.index * this.query.limit] ).limit( this.query.limit );
      break;
      default:
      break;
    }
    return fquery;
  }
  /// Set the query
  /// After setting the query, execute [reload()] to reflect the setting.
  FirestoreQuery query = FirestoreQuery( );

  /// Order for sorting
  OrderBy orderBy;
  /// Compare key for sorting
  String orderByKey;
  /// Order for sorting
  /// Apply when the first sort is equivalent
  OrderBy thenBy;
  /// Compare key for sorting
  /// Apply when the first sort is equivalent
  String thenByKey;

  /// Sort according to the sort setting
  /// [orderBy] and [orderByKey], [thenBy] and [thenByKey] as the sort contents.
  void sort( ) => this.data.sort( this._sortAction );
  int _sortAction( MapEntry<String, FirestoreDocument> a, MapEntry<String, FirestoreDocument> b ) {
    int r = 0;
    if( a.value != null && b.value != null && a.value.containsKey( this.orderByKey ) && b.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        if( a.value[this.orderByKey] is double ) {
          r = a.value.getDouble( this.orderByKey ).compareTo( b.value.getDouble( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is int ) {
          r = a.value.getInt( this.orderByKey ).compareTo( b.value.getInt( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is bool ) {
          r = a.value.getBool( this.orderByKey ).compareTo( b.value.getBool( this.orderByKey ) );
        } else {
          r = a.value.getString( this.orderByKey ).compareTo( b.value.getString( this.orderByKey ) );
        }
      } else {
        if( a.value[this.orderByKey] is double ) {
          r = b.value.getDouble( this.orderByKey ).compareTo( a.value.getDouble( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is int ) {
          r = b.value.getInt( this.orderByKey ).compareTo( a.value.getInt( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is bool ) {
          r = b.value.getBool( this.orderByKey ).compareTo( a.value.getBool( this.orderByKey ) );
        } else {
          r = b.value.getString( this.orderByKey ).compareTo( a.value.getString( this.orderByKey ) );
        }
      }
    } else if( a.value == null || !a.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        r = -1;
      } else {
        r = 1;
      }
    } else if( b.value == null || !b.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        r = 1;
      } else {
        r = -1;
      }
    }
    if( r == 0 && this.thenBy != OrderBy.None && !isEmpty( this.thenByKey ) ) {
      if( a.value != null && b.value != null && a.value.containsKey( this.thenByKey ) && b.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          if( a.value[this.thenByKey] is double ) {
            r = a.value.getDouble( this.thenByKey ).compareTo( b.value.getDouble( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is int ) {
            r = a.value.getInt( this.thenByKey ).compareTo( b.value.getInt( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is bool ) {
            r = a.value.getBool( this.thenByKey ).compareTo( b.value.getBool( this.thenByKey ) );
          } else {
            r = a.value.getString( this.thenByKey ).compareTo( b.value.getString( this.thenByKey ) );
          }
        } else {
          if( a.value[this.thenByKey] is double ) {
            r = b.value.getDouble( this.thenByKey ).compareTo( a.value.getDouble( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is int ) {
            r = b.value.getInt( this.thenByKey ).compareTo( a.value.getInt( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is bool ) {
            r = b.value.getBool( this.thenByKey ).compareTo( a.value.getBool( this.thenByKey ) );
          } else {
            r = b.value.getString( this.thenByKey ).compareTo( a.value.getString( this.thenByKey ) );
          }
        }
      } else if( a.value == null || !a.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          r = -1;
        } else {
          r = 1;
        }
      } else if( b.value == null || !b.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          r = 1;
        } else {
          r = -1;
        }
      }
    }
    return r;
  }

  /// True if communicating with the server for saving or deleting
  bool get isUpdating {
    for( FirestoreDocument doc in this.data.values ){
      if( doc == null ) continue;
      if( doc.isUpdating ) return true;
    }
    return false;
  }

  
  /// Get the protocol of the path
  @override
  String get protocol {
    if( isEmpty( this.rawPath.scheme ) ) return "firestore";
    return this.rawPath.scheme;
  }
  /// Destroys the object
  /// 
  /// Destroyed objects are not allowed
  @override  
  void dispose( ) {
    super.dispose( );
    this._disposeInternal( );
  }
  void _disposeInternal( ) {
    this.__reference = null;
    if( this._listener == null ) return;
    this._listener.cancel( );
    this._listener = null;
  }
  /// Callback event when application quit
  void onApplicationQuit( ) {
    super.onApplicationQuit( );
    this._disposeInternal( );
  }

}
